from . import err
from .err import err_enum
from .parser import web_parser
import requests
import pickle
import shutil
import time
import random
import os

g_web_acc_link = "https://store.steampowered.com/account/?l=english"
g_web_link = "https://store.steampowered.com/account/remotestorage/?l=english"
g_random_sleep_interval = (1, 3)

def random_sleep(func):
    def wrapper(*args, **kwargs):
        time.sleep(random.randint(g_random_sleep_interval[0], g_random_sleep_interval[1]))
        return func(*args, **kwargs)
    return wrapper

class web:
    def __init__(self, cookie):
        if not os.path.isfile(cookie):
            raise err.err(err_enum.NO_SESSION)
        self.web_parser = web_parser()
        self.session = requests.Session()
        self.cookie_pkl = cookie
        try:
            with open(self.cookie_pkl, 'rb') as f:
                self.session.cookies.update(pickle.load(f))
        except:
            raise err.err(err_enum.INVALID_COOKIE_FORMAT)

        response = self.session.get(g_web_acc_link)

    # Return list of {"Game": name, "Link", link}
    @random_sleep
    def get_list(self):
        response = self.session.get(g_web_link)
        if (response.status_code != 200):
            raise err(err_enum.ERR_CANNOT_RETRIEVE_LIST_MSG)

        return self.web_parser.parse_index(response.text)

    @random_sleep
    def _get_game_save(self, game_link:str):
        response = self.session.get(game_link)
        if (response.status_code != 200):
            err.err_enum(err_enum.CANNOT_RETRIEVE_GAME_FILES).log
            return (None, None)

        return self.web_parser.parse_game_file(response.text)


    def get_game_save(self, game_link:str):
        # Some games have a lot of games with multiple pages
        next_page_link = game_link
        save_file_list = list()
        while True:
            partial_save_file_list, next_page_link = \
                self._get_game_save(next_page_link)
            save_file_list += partial_save_file_list
            if next_page_link is None:
                break
        return save_file_list

    @random_sleep
    def download_game_save(self, link:str, store_location:str):
        with self.session.get(link, stream=True) as r:
            with open(store_location, 'wb') as f:
                    shutil.copyfileobj(r.raw, f)



from enum import Enum
from .err import err
from .err import err_enum
from . import ver
import logging
from discord_webhook import DiscordWebhook

logger = logging.getLogger(__name__)

class notify_method(Enum):
    Nop = 0
    Discord = 1

class notifier:

    '''
    Method: Nop (Do nothing)
    Method: Discord
        Required: webhook
    '''
    instance = None

    def get_enum(name:str):
        if name.lower() == 'discord':
            return notify_method.Discord
        else:
            return notify_method.Nop

    def is_supported(name:str):
        if len(name) == 0:
            return True
        enum_ = notifier.get_enum(name)
        if enum_ is None:
            return False
        else:
            return True

    def get_instance():
        return notifier.instance

    def create_instance(method:str, **kwargs):
        method_enum = notifier.get_enum(method)
        notifier.instance = notifier(method_enum, **kwargs)

        return notifier.instance

    def __init__(self, method:notify_method, **kwargs):
        self.method = method
        if (self.method == notify_method.Nop):
            pass
        elif (self.method == notify_method.Discord):
            assert 'webhook' in kwargs, 'Discord method requires webhook'
            self.webhook = kwargs['webhook']

            if len(self.webhook) == 0:
                logger.error(err.get_msg(err_enum.INVALID_WEBHOOK_URL))
                exit(1)
        else:
            assert False, 'Unsupported notifier method'

    def discord_send(self, msg:str):
        webhook = DiscordWebhook(url=self.webhook, content=msg)
        webhook = webhook.execute()

    def send(self, msg:str, ok:bool):
        actual_msg = f'[scsd-{ver.__version__}] '
        if ok:
            actual_msg += ":white_check_mark: "
        else:
            actual_msg += ":x: "
        actual_msg += f' {msg}'
        if self.method == notify_method.Nop:
            return
        elif self.method == notify_method.Discord:
            self.discord_send(actual_msg)
        else:
            assert False



import configparser
import os
import pathlib
from .err import err
from .err import err_enum
from .notifier import notifier, notify_method

Defaults = {
    'Required': {
    },
    'Rotation': {
        "rotation": 15
    },
    'Log': {
        "log_level": 2
    },
    'Notifier': {
        "notifier": "",
        "webhook": ""
    },
    'Target': {
        "mode": None,
        "list": None
    }
}

class config:
    def raise_err(self, additional_info=""):
        self.parse_error.set_additional_info(additional_info)
        raise self.parse_error
    def __init__(self, file=None):
        self.parse_error = err(err_enum.INVALID_CONFIG)

        self.config_file = file
        self.parser = configparser.ConfigParser()

        if self.config_file is not None:
            try:
                self.parser.read(self.config_file)
            except configparser.MissingSectionHeaderError as e:
                self.raise_err("[PLACEHOLDER_TEXT] section required for entire file")
        else:
            self.parser = None

        self.parsed = dict()
        for key in Defaults.keys():
            self.parsed[key] = dict()

    def is_file(self, arg) -> pathlib.Path:
        if os.path.isfile(arg):
            return pathlib.Path(arg)
        else:
            self.raise_err(f"File '{arg}' does not exist.")

    def is_path(self, arg) -> pathlib.Path:
        if os.path.isdir(arg):
            return pathlib.Path(arg)
        else:
            self.raise_err(f"Directory '{arg}' does not exist.")

    def check_and_raise(self, obj, entry:str):
        if entry not in obj:
            self.raise_err(f"'{entry}' required in config")

    def load_default(self, p_section:str, p_entry:str):
        self.parsed[p_section][p_entry] = Defaults[p_section][p_entry]

    def parse_required(self):
        if 'Required' not in self.parser:
            self.raise_err("No [Required] section found")
        required = self.parser['Required']

        self.check_and_raise(required, 'save_dir')
        self.parsed['Required']['save_dir'] = self.is_path(required['save_dir'])

    def parse_optional_section(self, p_section:str, p_entries:list):
        if self.parser is not None and p_section in self.parser:
            section = self.parser[p_section]
        else:
            section = dict()

        for entry in p_entries:
            if entry in section:
                self.parsed[p_section][entry] = section[entry]
            else:
                self.load_default(p_section, entry)

    def parse_log(self):
        self.parse_optional_section('Log', ['log_level'])

    def parse_rotation(self):
        self.parse_optional_section('Rotation', ['rotation'])

    def parse_notifier(self):
        self.parse_optional_section('Notifier', ['notifier', 'webhook'])
        if not notifier.is_supported(self.parsed['Notifier']['notifier']):
            self.raise_err(f"Unsupported notifier method '{self.parsed['Notifier']['notifier']}'")

    def delimit_list(self, p_input:str) -> list:
        if p_input is None:
            return None
        return [int(x) for x in p_input.strip().split(',')]

    def parse_target(self):
        self.parse_optional_section('Target', ['mode', 'list'])
        self.parsed['Target']['list'] = self.delimit_list(self.parsed['Target']['list'])

    def get_conf(self):
        self.parse_required()
        self.parse_log()
        self.parse_rotation()
        self.parse_notifier()
        self.parse_target()
        return self.parsed

    def load_from_arg(self, parsed_args:dict):
        self.parsed['Required']['save_dir'] = parsed_args['save_dir']
        self.parse_log()
        self.parsed['Log']['log_level'] = parsed_args['log_level']
        self.parse_rotation()
        self.parsed['Rotation']['rotation'] = parsed_args['rotation']
        self.parse_notifier()
        self.parse_target()
        if 'auth' in parsed_args:
            self.parsed['auth'] = parsed_args['auth']
        return self.parsed



from . import err
from .notifier import notifier
from .err import err_enum
from bs4 import BeautifulSoup
import datetime
import os
import logging

logger = logging.getLogger('scsd')

def get_tbody(soup):
    main_content = soup.find(id='main_content')

    if (main_content is None):
        raise err.err(err_enum.CANNOT_PARSE_LIST)

    if (not hasattr(main_content, "table")):
        raise err.err(err_enum.CANNOT_PARSE_LIST)

    return main_content.table.tbody

def parse_time(input:str) -> datetime.datetime:
    if '@' not in input:
        raise err.err(err_enum.CANNOT_PARSE_GAME_FILES)

    # Assume 'DD MMM [YYYY] @ HH:MM{a|p}m' format

    tokens = input.split(' ')
    datetime_ = None
    try:
        if len(tokens) == 4:
            now = datetime.datetime.now(tz=datetime.timezone.utc)
            year = now.year
            d = datetime.datetime.strptime(input, "%d %b @ %I:%M%p")
            datetime_ = d.replace(year=year, tzinfo=datetime.timezone.utc)
        elif len(tokens) == 5:
            datetime_ = datetime.datetime.strptime(input, "%d %b, %Y @ %I:%M%p").replace(tzinfo=datetime.timezone.utc)
        else:
            raise err.err(err_enum.CANNOT_PARSE_GAME_FILES)
    except ValueError:
        raise err.err(err_enum.CANNOT_PARSE_GAME_FILES)

    return datetime_

def get_appid(link:str) -> int:
    appid_token = 'appid='
    appid_location = link.find(appid_token)

    if (appid_location == -1):
        return -1

    return int(link[appid_location + len(appid_token):])

class web_parser:
    def __init__(self):
        pass

    def parse_index(self, content):
        try:
            return self._parse_index(content)
        except err.err as e:
            print('There are a few possibilities:\n 1. Your cookie has expired.\n 2. It seems like Steam has update the webpage. Please update to the latest version or notify the author.')
            raise e

    def _parse_index(self, content):
        soup = BeautifulSoup(content, 'html.parser')

        tbody = get_tbody(soup)

        data = list()

        rows = tbody.find_all('tr')
        for row in rows:
            cols = row.find_all('td')
            data.append({
                "name": cols[0].text.strip(),
                "link": cols[3].a['href'],
                "app_id": get_appid(cols[3].a['href'])
            })
        return data

    def parse_game_file(self, content) -> tuple:
        try:
            return self._parse_game_file(content)
        except err.err as e:
            e.print()
            exit(e.err_enum)

    def _parse_game_file(self, content) -> tuple:
        soup = BeautifulSoup(content, 'html.parser')

        tbody = get_tbody(soup)

        data = list()
        rows = tbody.find_all('tr')

        for row in rows:
            cols = row.find_all('td')
            path, filename = os.path.split(cols[1].text.strip())
            time_str = cols[3].text.strip()
            parsed_time = parse_time(time_str)
            logger.debug(f"Parse {filename} time '{time_str}' as '{parsed_time.isoformat()}'")
            data.append({
                "filename": filename,
                "path": path,
                "time": parsed_time,
                "link": cols[4].a['href']})

        has_next = soup.find('a', text='next >>')

        if (has_next is None):
            return (data, None)
        else:
            return (data, has_next['href'])



import sys
from enum import IntEnum
import logging

logger = logging.getLogger('scsd')

class err_enum(IntEnum):
    CANNOT_RETRIEVE_LIST = 1
    CANNOT_PARSE_LIST = 2
    CANNOT_RETRIEVE_GAME_FILES = 3
    CANNOT_PARSE_GAME_FILES = 4
    CANNOT_INITIALIZE_DB = 5
    CANNOT_CREATE_DIRECTORY = 6
    INVALID_WEBHOOK_URL = 7
    INVALID_COOKIE_FORMAT = 8,
    INVALID_CONFIG = 9,
    LOGIN_FAIL = 10,
    NO_SESSION = 11,
    CANNOT_REMOVE_OUTDATED = 12,
    REQUESTS_LIMIT_EXCEED = 13,
    UNKNOWN_EXCEPTION = 256

ERR_INFO = {
    err_enum.CANNOT_RETRIEVE_LIST: [
        logging.ERROR,
        "Cannot retrieve list from steam. Please make sure connected to Internet and cookie is valid."
    ],
    err_enum.CANNOT_PARSE_LIST: [
        logging.ERROR,
        "Cannot parse the list."
    ],
    err_enum.CANNOT_RETRIEVE_GAME_FILES: [
        logging.ERROR,
        "Cannot get game files."
    ],
    err_enum.CANNOT_PARSE_GAME_FILES: [
        logging.ERROR,
        "Cannot parse the file list. It seems like Steam has update the webpage. Please update to the latest version or notify the author."
    ],
    err_enum.CANNOT_INITIALIZE_DB: [
        logging.ERROR,
        "Cannot initialize database"
    ],
    err_enum.CANNOT_CREATE_DIRECTORY: [
        logging.ERROR,
        "Cannot create directory"
    ],
    err_enum.INVALID_WEBHOOK_URL: [
        logging.ERROR,
        "Invalid webhook url"
    ],
    err_enum.INVALID_COOKIE_FORMAT: [
        logging.ERROR,
        "Invalid cookie format. Should be Netscape format"
    ],
    err_enum.INVALID_CONFIG: [
        logging.ERROR,
        "Invalid config format: "
    ],
    err_enum.LOGIN_FAIL: [
        logging.ERROR,
        "Login fail, please try again"
    ],
    err_enum.NO_SESSION: [
        logging.ERROR,
        "No session found. Please re-run the program with -a <username> -d <data_dir> first"
    ],
    err_enum.CANNOT_REMOVE_OUTDATED: [
        logging.WARNING,
        "Cannot remove outdated file: "
    ],
    err_enum.REQUESTS_LIMIT_EXCEED: [
        logging.ERROR,
        "Request limit exceed. Please wait until another day"
    ],
    err_enum.UNKNOWN_EXCEPTION: [
        logging.ERROR,
        "Unknown exception"
    ]
}


class err(Exception):
    def log(self):
        msg = self.message
        if self.additional_info:
            msg += self.additional_info
        if self.level == logging.ERROR:
            logger.error(msg)
        elif self.level == logging.WARNING:
            logger.warning(msg)
        elif self.level == logging.INFO:
            logger.info(msg)
        else:
            logger.debug(msg)

    def get_msg(err_enum: err_enum) -> str:
        return ERR_INFO[err_enum][1]

    def get_msg(self) -> str:
        if self.additional_info is None:
            return self.message
        else:
            return self.message + self.additional_info

    def num(self):
        return self.err_enum.value

    def __init__(self, err_enum_:err_enum):
        self.err_enum = err_enum_
        self.message = ERR_INFO[self.err_enum][1]
        self.level = ERR_INFO[self.err_enum][0]
        self.additional_info = None
        super().__init__(self.message)

    def set_additional_info(self, info:str):
        self.additional_info = info



import sqlite3
import os
import datetime
from . import err
from .err import err_enum
import logging

DB_FILENAME = 'scsd.sqlite3'

'''Schema
TABLE GAMES
app_id (integer) PK, CONSTRAINT >0
game_name (text)
dir_name (text)

TABLE FILES
file_id (INT) PK
filename (TEXT)
path (TEXT)
app_id (INT)

TABLE VERSION
version_id (INT) PK AUTOINC
file_id foreign key
time (datetime)
version_num (INT) >= 0

TABLE REQUESTS
id (INT) PK NOT NULL DEFAULT 0
time (datetime)
count (int)
## Enforce single row


'''
logger = logging.getLogger('scsd')
class db:
    requests_limit = 85000
    def __init__(self, db_location:str, rotation:int):
        self.location = db_location
        self.rotation = rotation
        self.db_file = os.path.join(db_location, DB_FILENAME)

        self.con = sqlite3.connect(self.db_file, detect_types=sqlite3.PARSE_DECLTYPES)

        if not self.schema_ok():
            self.initialize_schema()

    def __del__(self):
        if hasattr(self, 'con'):
            self.con.close()

    def schema_ok(self) -> bool:
        cur = self.con.cursor()
        res = cur.execute("SELECT name FROM sqlite_master WHERE type='table' AND name='GAMES';")
        if res.fetchone() is None:
            return False

        res = cur.execute("SELECT name FROM sqlite_master WHERE type='table' AND name='FILES';")
        if res.fetchone() is None:
            return False

        res = cur.execute("SELECT name FROM sqlite_master WHERE type='table' AND name='REQUESTS';")
        if res.fetchone() is None:
            return False

        res = cur.execute("SELECT name FROM sqlite_master WHERE type='table' AND name='VERSION';")
        if res.fetchone() is None:
            return False

        return True

    def initialize_schema(self):
        cur = self.con.cursor()
        res = cur.execute(
            "CREATE TABLE GAMES("
            "  app_id INTEGER CHECK (app_id>0),"
            "  game_name text,"
            "  dir_name text,"
            "  PRIMARY KEY (app_id)"
            ");")

        res = cur.execute(
            "CREATE TABLE FILES("
            "  file_id INTEGER PRIMARY KEY AUTOINCREMENT,"
            "  filename TEXT,"
            "  path TEXT,"
            "  app_id INTEGER,"
            "  FOREIGN KEY (app_id) REFERENCES GAMES(app_id)"
            ");")

        res = cur.execute(
            "CREATE TABLE VERSION("
            "  version_id INTEGER PRIMARY KEY AUTOINCREMENT,"
            "  file_id INTEGER,"
            "  time timestamp,"
            "  version_num INTEGER CHECK (version_num>=0),"
            "  FOREIGN KEY (file_id) REFERENCES FILES(file_id)"
            ");")

        res = cur.execute(
            "CREATE TABLE REQUESTS("
            "  id INTEGER PRIMARY KEY CHECK (id = 0),"
            "  time timestamp,"
            "  num int DEFAULT 0"
            ");")

        res = cur.execute("INSERT INTO REQUESTS VALUES (?, ?, 0);", (0, datetime.datetime.now()))

        self.con.commit()

        if not self.schema_ok():
            raise err.err(err_enum.CANNOT_INITIALIZE_DB)

    def add_requests_count(self, count:int):
        cur = self.con.cursor()
        res = cur.execute("SELECT time FROM REQUESTS WHERE id = 0;");
        db_time = res.fetchone()[0]

        now = datetime.datetime.now()
        if (db_time.date() == now.date()):
            res = cur.execute("UPDATE REQUESTS SET num = num + ? WHERE id = 0;", (count,))
        else:
            res = cur.execute("UPDATE REQUESTS SET num = ?, time = ? WHERE id = 0;", (count, now))
        self.con.commit()

    def is_requests_limit_exceed(self) -> bool:
        cur = self.con.cursor()
        res = cur.execute("SELECT num FROM REQUESTS WHERE id = 0;")

        count = res.fetchone()[0]

        return count >= db.requests_limit

    def add_new_game(self, app_id:int, game_name:str):
        cur = self.con.cursor()
        res = cur.execute("INSERT INTO GAMES VALUES (?, ?, NULL);", (app_id, game_name))
        self.con.commit()

    def set_game_dir(self, app_id:int, dir_name:str):
        cur = self.con.cursor()
        res = cur.execute("UPDATE GAMES SET dir_name = ? WHERE app_id = ?;", (dir_name, app_id))
        self.con.commit()

    def get_game_dir(self, app_id:int) -> str:
        cur = self.con.cursor()
        res = cur.execute("SELECT dir_name FROM GAMES WHERE app_id = ?;", (app_id,))
        game_dir_tuple = res.fetchone()

        if game_dir_tuple is None:
            return None

        return game_dir_tuple[0]

    def get_file_id(self, app_id:int, filename:str) -> int:
        cur = self.con.cursor()
        res = cur.execute("SELECT file_id FROM FILES WHERE app_id=? AND filename=?;", (app_id, filename))
        file_id_tuple = res.fetchone()
        if file_id_tuple is None:
            return None
        return file_id_tuple[0]

    def is_game_exist(self, app_id:int):
        cur = self.con.cursor()
        res = cur.execute("SELECT app_id FROM GAMES WHERE app_id = ?;", (app_id,))
        return res.fetchone() is not None

    def get_game_files(self, app_id: int):
        pass

    # Expect [(filename, path, app_id, last_update_time), ...]
    def add_new_files(self, file_tuples:list):
        cur = self.con.cursor()
        for tup in file_tuples:
            res = cur.execute("INSERT INTO FILES VALUES (NULL, ?, ?, ?);", tup[0:3])
            file_id = cur.lastrowid

            no_tz_time = tup[3].replace(tzinfo=None)
            res = cur.execute("INSERT INTO VERSION VALUES (NULL, ?, ?, ?);", (file_id, no_tz_time, 0))

        self.con.commit()

    def is_file_outdated(self, file_id:int, server_time:datetime) -> bool:
        cur = self.con.cursor()
        res = cur.execute("SELECT time FROM VERSION WHERE file_id = ? and version_num = 0;", (file_id,))
        db_time_tuple = res.fetchone()
        if db_time_tuple is None:
            logger.warning(f'Failed to retrieve newest version time for file_id {file_id}')
            return True

        tz_db_time = db_time_tuple[0].replace(tzinfo=datetime.timezone.utc)

        logger.debug(f'DB time: {tz_db_time}, Server time: {server_time}')

        if tz_db_time < server_time:
            return True
        else:
            return False

    # +1 for each version
    # Insert 0 as now
    # Return max version
    def update_file_update_time_to_now(self, file_id:int, newest_file_time: datetime.datetime) -> int:
        cur = self.con.cursor()

        # Increment one for all
        res = cur.execute("UPDATE VERSION SET version_num = version_num + 1 WHERE file_id = ?;", (file_id,))

        time_without_tz = newest_file_time.replace(tzinfo=None)
        # Insert newest as 0
        res = cur.execute("INSERT INTO VERSION VALUES (NULL, ?, ?, 0)", (file_id, time_without_tz))
        self.con.commit()

        res = cur.execute("SELECT COUNT(*) FROM VERSION WHERE file_id = ?", (file_id,))
        count = res.fetchone();

        if (count is None):
            return 0
        else:
            return count[0]

    def remove_outdated_file(self, file_id:int):
        cur = self.con.cursor()
        res = cur.execute("SELECT version_num FROM VERSION WHERE file_id = ? AND version_num >= ?", (file_id, self.rotation))
        outdated_version_num = res.fetchall();

        res = cur.execute("DELETE FROM VERSION WHERE file_id = ? AND version_num >= ?", (file_id, self.rotation))

        self.con.commit()

        logger.debug(f"DB Removing version {outdated_version_num}")

        return outdated_version_num
